/*
 * Name:    d_epmcs.c 
 *
 * Purpose: MX multichannel scaler driver for the EPICS multichannel
 *          scaler support.  This currently is just for the Struck MCS.
 *
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 *
 * Warning: This driver is just a stub for the use of the Struck MCS
 *          EPICS quick scan support.  It does not yet support use of
 *          the MCS via an mx_mcs_... API.  This will come later.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_mcs.h"
#include "d_epmcs.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the mcs driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_mcs_record_function_list = {
	mxd_epics_mcs_initialize_type,
	mxd_epics_mcs_create_record_structures,
	mxd_epics_mcs_finish_record_initialization,
	mxd_epics_mcs_delete_record,
	NULL,
	mxd_epics_mcs_read_parms_from_hardware,
	mxd_epics_mcs_write_parms_to_hardware,
	mxd_epics_mcs_open,
	mxd_epics_mcs_close
};

/* EPICS mcs data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_mcs_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXD_EPICS_MCS_STANDARD_FIELDS
};

long mxd_epics_mcs_num_record_fields
		= sizeof( mxd_epics_mcs_record_field_defaults )
			/ sizeof( mxd_epics_mcs_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_mcs_rfield_def_ptr
			= &mxd_epics_mcs_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_epics_mcs_get_pointers( MX_MCS *mcs,
			MX_EPICS_MCS **epics_mcs,
			const char *calling_fname )
{
	const char fname[] = "mxd_epics_mcs_get_pointers()";

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCS pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( epics_mcs == (MX_EPICS_MCS **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_MCS pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( mcs->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_MCS pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*epics_mcs = (MX_EPICS_MCS *) mcs->record->record_type_struct;

	if ( *epics_mcs == (MX_EPICS_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_EPICS_MCS pointer for record '%s' is NULL.",
			mcs->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_epics_mcs_initialize_type( long type )
{
		return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_mcs_create_record_structures()";

	MX_MCS *mcs;
	MX_EPICS_MCS *epics_mcs;

	/* Allocate memory for the necessary structures. */

	mcs = (MX_MCS *) malloc( sizeof(MX_MCS) );

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCS structure." );
	}

	epics_mcs = (MX_EPICS_MCS *) malloc( sizeof(MX_EPICS_MCS) );

	if ( epics_mcs == (MX_EPICS_MCS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_MCS structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = mcs;
	record->record_type_struct = epics_mcs;
	record->class_specific_function_list = NULL;

	mcs->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_finish_record_initialization( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_open( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_mcs_open()";

	MX_MCS *mcs;
	MX_EPICS_MCS *epics_mcs;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	char epics_device_type[80];
	int ezca_status;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	mcs = (MX_MCS *) (record->record_class_struct);

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	/* Here we find out if this EPICS record really exists in an
	 * IOC's database by asking for its EPICS device type.
	 */

	strncpy( pvname, epics_mcs->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".DTYP" );

	ezca_status = ezcaGet( pvname, ezcaString, 1, epics_device_type );

	MX_DEBUG( 2, ("(%s) epics device type = '%s'",
		epics_mcs->epics_record_name, epics_device_type));

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error getting EPICS device type for MX record '%s'. EZCA status = %d (%s)",
			record->name, ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

