/*
 * Name:    d_pmac.c
 *
 * Purpose: MX driver for Delta Tau PMAC controlled motors.
 *
 *          Note that this driver performs its work using the PMAC jog
 *          commands, rather than using a motion program.  The expectation
 *          is that this is less likely to interfere with any running
 *          motor programs.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "d_pmac.h"

/* ============ Motor channels ============ */

MX_RECORD_FUNCTION_LIST mxd_pmac_record_function_list = {
	mxd_pmac_initialize_type,
	mxd_pmac_create_record_structures,
	mxd_pmac_finish_record_initialization,
	mxd_pmac_delete_record,
	mxd_pmac_print_structure,
	mxd_pmac_read_parms_from_hardware,
	mxd_pmac_write_parms_to_hardware,
	mxd_pmac_open,
	mxd_pmac_close
};

MX_MOTOR_FUNCTION_LIST mxd_pmac_motor_function_list = {
	mxd_pmac_motor_is_busy,
	mxd_pmac_move_absolute,
	mxd_pmac_get_position,
	mxd_pmac_set_position,
	mxd_pmac_soft_abort,
	mxd_pmac_immediate_abort,
	mxd_pmac_positive_limit_hit,
	mxd_pmac_negative_limit_hit,
	mxd_pmac_find_home_position,
};

#define PMAC_DEBUG	FALSE

MX_RECORD_FIELD_DEFAULTS mxd_pmac_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_STEPPER_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_PMAC_STANDARD_FIELDS
};

long mxd_pmac_num_record_fields
		= sizeof( mxd_pmac_record_field_defaults )
			/ sizeof( mxd_pmac_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_pmac_rfield_def_ptr
			= &mxd_pmac_record_field_defaults[0];

MX_EXPORT mx_status_type
mxd_pmac_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_create_record_structures()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) malloc( sizeof(MX_PMAC_MOTOR) );

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_PMAC_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = pmac_motor;
	record->class_specific_function_list
				= &mxd_pmac_motor_function_list;

	motor->record = record;

	/* A PMAC motor is treated as an stepper motor. */

	motor->subclass = MXC_MTR_STEPPER;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_finish_record_initialization()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	MX_PMAC *pmac;
	double delay_seconds;
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	/*** Server minimum delay time between commands. ***/

	delay_seconds = 0.0;

	record->event_interval
		= mx_convert_seconds_to_clock_ticks( delay_seconds );

	pmac_motor = (MX_PMAC_MOTOR *) record->record_type_struct;

	if ( ( pmac_motor->motor_number < 1 )
				|| ( pmac_motor->motor_number > 8 ) )
	{
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Illegal PMAC motor number %d for record '%s'.  Allowed range is 1 to 8.",
			pmac_motor->motor_number, record->name );
	}

	/* Save a pointer to the motor's MX_MOTOR structure in the MX_PMAC
	 * structure, so that the eight motors can find each other.
	 */

	motor = (MX_MOTOR *) record->record_class_struct;

	pmac = (MX_PMAC *)
		pmac_motor->pmac_interface.record->record_type_struct;

	pmac->motor[ pmac_motor->motor_number - 1 ] = motor;

	return status;
}

MX_EXPORT mx_status_type
mxd_pmac_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_print_structure()";

	MX_MOTOR *motor;
	MX_INTERFACE *pmac_interface;
	MX_PMAC_MOTOR *pmac_motor;
	long motor_steps;
	double position, backlash, negative_limit, positive_limit, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_PMAC_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	pmac_interface = &(pmac_motor->pmac_interface);

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);

	fprintf(file, "  Motor type     = PMAC motor.\n\n");

	fprintf(file, "  name           = %s\n", record->name);
	fprintf(file, "  port name      = %s\n", pmac_interface->record->name);
	fprintf(file, "  motor number   = %d\n", pmac_motor->motor_number);

	status = mx_motor_get_position_steps( record, &motor_steps );

	if ( status.code != MXE_SUCCESS ) {
		mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
				record->name );
	}
	
	position = motor->offset + motor->scale
			* (double) motor_steps;
	
	fprintf(file, "  position       = %ld steps (%g %s)\n",
			motor_steps, position, motor->units);
	fprintf(file, "  scale          = %g %s per step.\n",
			motor->scale, motor->units);
	fprintf(file, "  offset         = %g %s.\n",
			motor->offset, motor->units);
	
	backlash = motor->scale
			* (double) (motor->raw_backlash_correction.stepper);
	
	fprintf(file, "  backlash       = %ld steps (%g %s)\n",
			motor->raw_backlash_correction.stepper,
			backlash, motor->units);
	
	negative_limit = motor->offset
	+ motor->scale * (double)(motor->raw_negative_limit.stepper);
	
	fprintf(file, "  negative limit = %ld steps (%g %s)\n",
			motor->raw_negative_limit.stepper,
			negative_limit, motor->units);

	positive_limit = motor->offset
	+ motor->scale * (double)(motor->raw_positive_limit.stepper);

	fprintf(file, "  positive limit = %ld steps (%g %s)\n",
			motor->raw_positive_limit.stepper,
			positive_limit, motor->units);

	deadband = motor->scale * (double) motor->deadband.stepper;

	fprintf(file, "  deadband       = %ld steps (%g %s)\n\n",
			motor->deadband.stepper,
			deadband, motor->units );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_read_parms_from_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_read_parms_from_hardware()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	MX_GENERIC *generic;
	double position;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	generic = (MX_GENERIC *)
		pmac_motor->pmac_interface.record->record_class_struct;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_GENERIC pointer for motor record '%s' is NULL.",
			record->name );
	}

	/* Flush out any unwritten and unread characters before sending
	 * commands to the PMAC controller.
	 */

	status = mx_generic_discard_unwritten_output( generic, PMAC_DEBUG );

	if ( status.code != MXE_SUCCESS && status.code != MXE_UNSUPPORTED )
		return status;

	status = mx_generic_discard_unread_input( generic, PMAC_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the current position.  This has the side effect of
	 * updating the position value in the MX_MOTOR structure.
	 */

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* At present, the position is the only parameter we read out.
	 * We rely on the non-volatile memory of the controller
	 * for the rest of the controller parameters.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_write_parms_to_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_write_parms_to_hardware()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	MX_GENERIC *generic;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE,
			"MX_PMAC_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	generic = (MX_GENERIC *)
		pmac_motor->pmac_interface.record->record_class_struct;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_GENERIC pointer for record '%s' is NULL.",
			record->name );
	}

	/* Flush out any written and unread characters before sending
	 * commands to the PMAC controller.
	 */

	status = mx_generic_discard_unwritten_output( generic, PMAC_DEBUG );

	if ( status.code != MXE_SUCCESS && status.code != MXE_UNSUPPORTED )
		return status;

	status = mx_generic_discard_unread_input( generic, PMAC_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* === Now we can set the motor position. === */

#if 0	/*  For now, we do not do this. */

	status = mxd_pmac_set_position_steps(
			motor, motor->u.stepper.motor_steps );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* At present, the position is the only parameter we set here.
	 * We rely on the non-volatile memory of the controller
	 * for the rest of the controller parameters.
	 */
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_open( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_close( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

/* ============ Motor specific functions ============ */

MX_EXPORT mx_status_type
mxd_pmac_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_motor_is_busy()";

	MX_PMAC *pmac;
	MX_PMAC_MOTOR *pmac_motor;
	char command[20];
	char response[20];
	int length;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_MOTOR pointer for motor is NULL." );
	}

	pmac = (MX_PMAC *)
		pmac_motor->pmac_interface.record->record_type_struct;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC pointer for motor is NULL." );
	}

	sprintf( command, "#%d?", pmac_motor->motor_number );

	status = mxi_pmac_command( pmac, command,
			response, sizeof response, NULL, 0, PMAC_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length < 12 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Too few characters seen in response to motor status command.  "
	"Only saw %d characters.", length );
	}

	if ( (response[2] & 0x2) || (response[11] & 0x1) ) {
		motor->busy = FALSE;
	} else {
		motor->busy = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_move_absolute()";

	MX_PMAC *pmac;
	MX_PMAC_MOTOR *pmac_motor;
	char command[20];
	long motor_steps;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_MOTOR pointer for motor is NULL." );
	}

	pmac = (MX_PMAC *)
		pmac_motor->pmac_interface.record->record_type_struct;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC pointer for motor is NULL." );
	}

	motor_steps = motor->raw_destination.stepper;

	if (  (pmac->major_version > 1)
	 || ( (pmac->major_version == 1) && (pmac->minor_version >= 16) ) )
	{
		sprintf( command, "#%dJ==%ld",
				pmac_motor->motor_number, motor_steps );
	} else {
		sprintf( command, "#%dJ=%ld",
				pmac_motor->motor_number, motor_steps );
	}

	status = mxi_pmac_command( pmac, command,
					NULL, 0, NULL, 0, PMAC_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_pmac_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_get_position_steps()";

	MX_PMAC *pmac;
	MX_PMAC_MOTOR *pmac_motor;
	char command[20];
	char response[50];
	int num_tokens;
	double double_value;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_MOTOR pointer for motor is NULL." );
	}

	pmac = (MX_PMAC *)
		pmac_motor->pmac_interface.record->record_type_struct;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC pointer for motor is NULL." );
	}

	sprintf( command, "#%dP", pmac_motor->motor_number );

	status = mxi_pmac_command( pmac, command,
			response, sizeof response, NULL, 0, PMAC_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* The PMAC may report values back in terms of fractions of
	 * an encoder count.
	 */

	num_tokens = sscanf( response, "%lg", &double_value );

	if ( num_tokens != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"No position value seen in response to 'TP' command.  "
		"Response seen = '%s'", response );
	}

	/* We round this to the nearest encoder count. */

	motor->raw_position.stepper = mx_round( double_value );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_set_position()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
"Setting the current position value of a PMAC motor is not available yet." );
}

MX_EXPORT mx_status_type
mxd_pmac_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_soft_abort()";

	MX_PMAC *pmac;
	MX_PMAC_MOTOR *pmac_motor;
	char command[20];
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_MOTOR pointer for motor is NULL." );
	}

	pmac = (MX_PMAC *)
		pmac_motor->pmac_interface.record->record_type_struct;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC pointer for motor is NULL." );
	}

	sprintf( command, "#%dJ/", pmac_motor->motor_number );

	status = mxi_pmac_command( pmac, command,
					NULL, 0, NULL, 0, PMAC_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_pmac_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_immediate_abort()";

	MX_PMAC *pmac;
	MX_PMAC_MOTOR *pmac_motor;
	char command[20];
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_MOTOR pointer for motor is NULL." );
	}

	pmac = (MX_PMAC *)
		pmac_motor->pmac_interface.record->record_type_struct;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC pointer for motor is NULL." );
	}

	sprintf( command, "#%dJ/", pmac_motor->motor_number );

	status = mxi_pmac_command( pmac, command,
						NULL, 0, NULL, 0, PMAC_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_pmac_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_positive_limit_hit()";

	MX_PMAC *pmac;
	MX_PMAC_MOTOR *pmac_motor;
	char command[20];
	char response[20];
	int length;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_MOTOR pointer for motor is NULL." );
	}

	pmac = (MX_PMAC *)
		pmac_motor->pmac_interface.record->record_type_struct;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC pointer for motor is NULL." );
	}

	sprintf( command, "#%d?", pmac_motor->motor_number );

	status = mxi_pmac_command( pmac, command,
			response, sizeof response, NULL, 0, PMAC_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length < 12 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Too few characters seen in response to motor status command.  "
	"Only saw %d characters.", length );
	}

	if ( (response[0] & 0x2) == 0 ) {
		motor->positive_limit_hit = FALSE;
	} else {
		motor->positive_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_negative_limit_hit()";

	MX_PMAC *pmac;
	MX_PMAC_MOTOR *pmac_motor;
	char command[20];
	char response[20];
	int length;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_MOTOR pointer for motor is NULL." );
	}

	pmac = (MX_PMAC *)
		pmac_motor->pmac_interface.record->record_type_struct;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC pointer for motor is NULL." );
	}

	sprintf( command, "#%d?", pmac_motor->motor_number );

	status = mxi_pmac_command( pmac, command,
			response, sizeof response, NULL, 0, PMAC_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length < 12 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Too few characters seen in response to motor status command.  "
	"Only saw %d characters.", length );
	}

	if ( (response[0] & 0x4) == 0 ) {
		motor->negative_limit_hit = FALSE;
	} else {
		motor->negative_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_find_home_position()";

	MX_PMAC *pmac;
	MX_PMAC_MOTOR *pmac_motor;
	char command[20];
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_MOTOR pointer for motor is NULL." );
	}

	pmac = (MX_PMAC *)
		pmac_motor->pmac_interface.record->record_type_struct;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC pointer for motor is NULL." );
	}

	sprintf( command, "#%dHM", pmac_motor->motor_number );

	status = mxi_pmac_command( pmac, command,
					NULL, 0, NULL, 0, PMAC_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

