/*
 * Name:    i_ni488.c
 *
 * Purpose: MX driver for National Instruments GPIB interfaces.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_NI488

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#include "mxunistd.h"

#include "mx_util.h"
#include "mx_gpib.h"
#include "mx_rec.h"
#include "i_ni488.h"

/* Include the National Instruments driver include file. */

#ifdef OS_LINUX
#include "ib.h"
#else
#include "decl.h"
#endif

MX_RECORD_FUNCTION_LIST mxi_ni488_record_function_list = {
	mxi_ni488_initialize_type,
	mxi_ni488_create_record_structures,
	mxi_ni488_finish_record_initialization,
	mxi_ni488_delete_record,
	mxi_ni488_print_interface_structure,
	mxi_ni488_read_parms_from_hardware,
	mxi_ni488_write_parms_to_hardware,
	mxi_ni488_open,
	mxi_ni488_close
};

MX_GPIB_FUNCTION_LIST mxi_ni488_gpib_function_list = {
	mxi_ni488_open_device,
	mxi_ni488_close_device,
	mxi_ni488_read_parms_from_device,
	mxi_ni488_write_parms_to_device,
	mxi_ni488_read,
	mxi_ni488_write,
	mxi_ni488_interface_clear,
	mxi_ni488_device_clear,
	mxi_ni488_selective_device_clear,
	mxi_ni488_local_lockout,
	mxi_ni488_remote_enable,
	mxi_ni488_go_to_local,
	mxi_ni488_trigger,
	mxi_ni488_wait_for_service_request,
	mxi_ni488_serial_poll,
	mxi_ni488_serial_poll_disable
};

MX_RECORD_FIELD_DEFAULTS mxi_ni488_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_GPIB_STANDARD_FIELDS,
	MXI_NI488_STANDARD_FIELDS
};

long mxi_ni488_num_record_fields
		= sizeof( mxi_ni488_record_field_defaults )
			/ sizeof( mxi_ni488_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_ni488_rfield_def_ptr
			= &mxi_ni488_record_field_defaults[0];

/* ==== Private function for the driver's use only. ==== */

static char *
mxi_ni488_gpib_error_text( void )
{
	static char ibsta_not_set[] = 
	"Error flag in 'ibsta' is not set.  Are you sure there was an error?";

	static char error_message_table[][80] = {

	"EDVR - Operating system error",			/*  0 */
	"ECIC - Function requires GPIB board to be CIC",	/*  1 */
	"ENOL - No listeners on the GPIB bus",			/*  2 */
	"EADR - GPIB board not addressed correctly",		/*  3 */
	"EARG - Invalid argument to function call",		/*  4 */
	"ESAC - GPIB board not system controller as required",	/*  5 */
	"EABO - I/O operation aborted (timeout)",		/*  6 */
	"ENEB - Nonexistent GPIB board",			/*  7 */
	"EDMA - DMA error",					/*  8 */
	"EBTO - DMA hardware microprocessor bus timeout",	/*  9 Linux */
	"EOIP - New I/O attempted with old I/O in progress",	/* 10 */
	"ECAP - No capability for intended operation",		/* 11 */
	"EFSO - File system error",				/* 12 */
	"EOWN - Shareable board exclusively owned",		/* 13 Linux */
	"EBUS - GPIB bus error",				/* 14 */
	"ESTB - Serial poll status byte queue overflow",	/* 15 */
	"ESRQ - SRQ stuck in ON position",			/* 16 */
	"ECFG - Config file not found",				/* 17 Linux */
	"EPAR - Parse error in config file",			/* 18 Linux */
#ifdef OS_LINUX
	"ETAB - Table overflow",				/* 19 Linux */
	"ENSD - Device not found",				/* 20 Linux */
#else
	"E\?\?\? - Unknown error code",				/* 19 */
	"ETAB - Table problem",					/* 20 */
#endif
	"ENWE - Network error",					/* 21 Linux */
	"ENTF - Nethandle table full",				/* 22 Linux */
	"EMEM - Memory allocation error",			/* 23 Linux */
	"E\?\?\? - Unknown error code"
	};

	static int num_error_messages = sizeof( error_message_table )
					/ sizeof( error_message_table[0] );

	if ( (ibsta & ERR) == 0 ) {
		return &ibsta_not_set[0];
	}

	if ( iberr < 0 || iberr >= num_error_messages ) {
		return &error_message_table[ num_error_messages ][0];
	} else {
		return &error_message_table[ iberr ][0];
	}
}

/* ==== Public functions ==== */

MX_EXPORT mx_status_type
mxi_ni488_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_ni488_create_record_structures()";

	MX_GPIB *gpib;
	MX_NI488 *ni488;

	/* Allocate memory for the necessary structures. */

	gpib = (MX_GPIB *) malloc( sizeof(MX_GPIB) );

	if ( gpib == (MX_GPIB *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_GPIB structure." );
	}

	ni488 = (MX_NI488 *) malloc( sizeof(MX_NI488) );

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_NI488 structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = gpib;
	record->record_type_struct = ni488;
	record->class_specific_function_list
				= &mxi_ni488_gpib_function_list;

	gpib->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_finish_record_initialization( MX_RECORD *record )
{
	return mx_gpib_finish_record_initialization( record );
}

MX_EXPORT mx_status_type
mxi_ni488_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_print_interface_structure( FILE *file, MX_RECORD *record )
{
	MX_GPIB *gpib;
	MX_NI488 *ni488;
	char read_eos_char, write_eos_char;

	gpib = (MX_GPIB *) (record->record_class_struct);

	ni488 = (MX_NI488 *) (record->record_type_struct);

	fprintf(file, "GPIB parameters for interface '%s':\n", record->name);
	fprintf(file, "  GPIB type              = NI488.\n\n");

	fprintf(file, "  name                   = %s\n", record->name);
	fprintf(file, "  interface name         = %s\n", ni488->interface_name);
	fprintf(file, "  default I/O timeout    = %d\n",
					gpib->default_io_timeout_value);
	fprintf(file, "  default EOI mode       = %d\n",
					gpib->default_eoi_mode);

	read_eos_char = gpib->default_read_terminator;

	if ( isprint(read_eos_char) ) {
		fprintf(file,
		      "  default read EOS char  = 0x%02x (%c)\n",
			read_eos_char, read_eos_char);
	} else {
		fprintf(file,
		      "  default read EOS char  = 0x%02x\n", read_eos_char);
	}

	write_eos_char = gpib->default_write_terminator;

	if ( isprint(write_eos_char) ) {
		fprintf(file,
		      "  default write EOS char = 0x%02x (%c)\n",
			write_eos_char, write_eos_char);
	} else {
		fprintf(file,
		      "  default write EOS char = 0x%02x\n", write_eos_char);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_read_parms_from_hardware( MX_RECORD *record )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_write_parms_to_hardware( MX_RECORD *record )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_open( MX_RECORD *record )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_close( MX_RECORD *record )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

/* ========== Device specific calls ========== */

MX_EXPORT mx_status_type
mxi_ni488_open_device( MX_GPIB *gpib, int address )
{
	const char fname[] = "mxi_ni488_open_device()";

	MX_NI488 *ni488;
	int descriptor;

#if defined( OS_LINUX )
	char device_name[40];
#else
	int board_number, num_items;
	unsigned long eos_value, read_value, write_value;
#endif

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

#ifdef OS_LINUX
	sprintf( device_name, "%s.%d", ni488->interface_name, address );

	MX_DEBUG( 2,("%s: Opening device '%s'.", fname, device_name));

	descriptor = ibfind( device_name );
#else
	num_items = sscanf( ni488->interface_name, "gpib%d", &board_number );

	if ( num_items != 1 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Board number not found in GPIB interface name '%s' for record '%s'",
			ni488->interface_name,
			gpib->record->name );
	}
	
	read_value = gpib->read_terminator[address];
	write_value = gpib->write_terminator[address];

	if ( read_value ) {
		eos_value = REOS | read_value;
	} else {
		eos_value = write_value;
	}

	if ( write_value ) {
		eos_value |= XEOS;
	}

	descriptor = ibdev( board_number, address, 0,
				gpib->eoi_mode[address], (int) eos_value );
#endif
	MX_DEBUG( 2,("%s: descriptor = %d", fname, descriptor));

	ni488->device_descriptor[address] = descriptor;

#ifdef OS_LINUX
	if ( descriptor == 32768 || descriptor == -1 ) {
#else
	if ( descriptor == -1 ) {
#endif
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Cannot open GPIB address %d on interface '%s'.  GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_close_device( MX_GPIB *gpib, int address )
{
	const char fname[] = "mxi_ni488_close_device()";

	MX_NI488 *ni488;
	int gpib_status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	gpib_status = ibonl( ni488->device_descriptor[address], 0 );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Cannot close GPIB address %d on interface '%s'.  GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_read_parms_from_device( MX_GPIB *gpib, int address )
{
	const char fname[] = "mxi_ni488_read_parms_from_device()";

	MX_NI488 *ni488;
	int dev, gpib_status, value;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

#if defined( OS_LINUX )
	/* Linux GPIB doesn't have ibask(), so we can't do anything here. */

	dev = gpib_status = value = 0;		/* Keep GCC happy. */

	return MX_SUCCESSFUL_RESULT;

#else /* OS_LINUX */

	/* Read all of the configuration parameters using ibask(). */

	dev = ni488->device_descriptor[address];

	/* Get secondary GPIB address. */

	gpib_status = ibask( dev, IbaSAD, &value );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot read secondary address for primary GPIB address %d on interface '%s'."
"  GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	gpib->secondary_address[address] = value;

	/* Get I/O timeout value. */

	gpib_status = ibask( dev, IbaTMO, &value );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot read I/O timeout value for GPIB address %d on interface '%s'.  "
"GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	gpib->io_timeout_value[address] = value;

	/* Get EOI mode. */

	gpib_status = ibask( dev, IbaEOT, &value );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot read EOI handling mode for GPIB address %d on interface '%s'.  "
"GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	gpib_device->eoi_mode[address] = value;

	/* Get read EOS character. */

	gpib_status = ibask( dev, IbaEOSrd, &value );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot get read EOS character for GPIB address %d on interface '%s'.  "
"GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	gpib_device->read_terminator[address] = (char) value;

	/* Get write EOS character. */

	gpib_status = ibask( dev, IbaEOSwrt, &value );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot get write EOS character for GPIB address %d on interface '%s'.  "
"GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	gpib_device->write_terminator[address] = (char) value;

	return MX_SUCCESSFUL_RESULT;

#endif /* OS_LINUX */
}

MX_EXPORT mx_status_type
mxi_ni488_write_parms_to_device( MX_GPIB *gpib, int address )
{
	const char fname[] = "mxi_ni488_write_parms_to_device()";

	MX_NI488 *ni488;
	int dev, gpib_status;
	unsigned long value, read_value, write_value;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	dev = ni488->device_descriptor[address];

	/* Set the I/O timeout value. */

	gpib_status = ibtmo( dev, gpib->io_timeout_value[address] );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot set I/O timeout value to '%d' for GPIB address %d on interface '%s'."
"  GPIB error = '%s'",
			gpib->io_timeout_value[address],
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	/* Set the EOI mode. */

	gpib_status = ibeot( dev, gpib->eoi_mode[address] );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot set EOI mode to '%d' for GPIB address %d on interface '%s'.  "
"GPIB error = '%s'",
			gpib->eoi_mode[address],
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	/* Setup the EOS mode and character. */

	MX_DEBUG( 2,("%s: must check for appropriate handling of EOS",fname));

	read_value = gpib->read_terminator[address];
	write_value = gpib->write_terminator[address];

	MX_DEBUG( 2,("%s: read_value = %d, write_value = %d",
			fname, read_value, write_value));

	if ( read_value != write_value ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Read and write EOS terminators must be the same for the NI-488 driver." );
	}

	if ( read_value > 0xff ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"EOS terminators for the NI-488 driver must be only one byte long." );
	}

	/* The Linux GPIB package seems to define the values of REOS and XEOS
	 * in /usr/local/include/gpib/gpib_user.h in a manner that is
	 * inconsistent with the National Instruments header file "decl.h"
	 * and also the National Instruments printed documentation.
	 * _However_, the value passed to the ibeos() function in
	 * the Linux driver function "driver/protocol/ibutil.c" is used
	 * in a way that _is_ compatible with the National Intruments
	 * header file and documentation.  Looks like a bug.  It must
	 * be that people on Linux are always setting the EOS modes
	 * in /etc/gpib.conf.  In any case, I have chosen to explicitly
	 * specify the bit patterns here rather than using the REOS and XEOS
	 * defines.
	 */

	if ( read_value ) {
		value = ( 0x4 << 8 ) | read_value;  /* should be REOS */
	} else {
		value = write_value;
	}

	if ( write_value ) {
		value |= ( 0x8 << 8 );             /* should be XEOS */
	}

	MX_DEBUG( 2,("%s: ibeos value = %#lx (%ld)", fname, value, value));

	gpib_status = ibeos( dev, (int) value );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot set EOS mode to '0x%x' for GPIB address %d on interface '%s'.  "
"GPIB error = '%s'",
			value,
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_read( MX_GPIB *gpib,
		int address,
		char *buffer,
		size_t max_bytes_to_read,
		size_t *bytes_read,
		int flags )
{
	const char fname[] = "mxi_ni488_read()";

	MX_NI488 *ni488;
	char *ptr;
	int dev, gpib_status;

	MX_DEBUG(-2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	dev = ni488->device_descriptor[address];

	if ( flags & MXF_GPIB_NOWAIT ) { 	/* Asynchronous read. */

#ifdef OS_LINUX
		return mx_error( MXE_UNSUPPORTED, fname,
	"Asynchronous reads are not supported under the Linux GPIB driver.");
#else 
		gpib_status = ibrda( dev, buffer, max_bytes_to_read );
#endif
	} else {				/* Synchronous read. */

		gpib_status = ibrd( dev, buffer, max_bytes_to_read );
	}

	if ( bytes_read != NULL ) {
#ifdef OS_LINUX
		*bytes_read = (size_t) ibcnt;
#else
		*bytes_read = (size_t) ibcntl;
#endif
	}

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Read for GPIB address %d on interface '%s' was unsuccessful.  "
"GPIB error = '%s'",
		address, ni488->interface_name,
		mxi_ni488_gpib_error_text() );
	}

	/* If it is present, strip off a line terminator at the end
	 * of the buffer.
	 */

	ptr = &buffer[ *bytes_read - 1 ];

	if ( *ptr == (char) gpib->read_terminator[ address ] ) {
		*ptr = '\0';

		(*bytes_read) --;
	}

#if 1
	MX_DEBUG(-2,("%s: dev = %d, *bytes_read = %ld, buffer = '%s'",
		fname, dev, *bytes_read, buffer));
#endif
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_write( MX_GPIB *gpib,
		int address,
		char *buffer,
		size_t bytes_to_write,
		size_t *bytes_written,
		int flags )
{
	const char fname[] = "mxi_ni488_write()";

	MX_NI488 *ni488;
	int dev, gpib_status;

	MX_DEBUG(-2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	dev = ni488->device_descriptor[address];

	if ( flags & MXF_GPIB_NOWAIT ) { 	/* Asynchronous write. */

#ifdef OS_LINUX
		return mx_error( MXE_UNSUPPORTED, fname,
	"Asynchronous writes are not supported under the Linux GPIB driver.");
#else
		gpib_status = ibwrta( dev, buffer, bytes_to_write );
#endif
	} else {				/* Synchronous write. */

	MX_DEBUG(-2,("%s: dev = %d, bytes_to_write = %ld, buffer = '%s'",
		fname, dev, bytes_to_write, buffer));

		gpib_status = ibwrt( dev, buffer, bytes_to_write );
	}
	if ( bytes_written != NULL ) {
#ifdef OS_LINUX
		*bytes_written = (size_t) ibcnt;
#else
		*bytes_written = (size_t) ibcntl;
#endif
	}
	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Write for GPIB address %d on interface '%s' was unsuccessful.  "
"GPIB error = '%s'",
		address, ni488->interface_name,
		mxi_ni488_gpib_error_text() );
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_interface_clear( MX_GPIB *gpib )
{
	const char fname[] = "mxi_ni488_interface_clear()";

	MX_NI488 *ni488;
	int gpib_status, interface_address;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	interface_address = ni488->interface_address;

	gpib_status = ibsic( ni488->device_descriptor[interface_address] );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot do an interface clear for GPIB interface '%s'.  GPIB error = '%s'",
			ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_device_clear( MX_GPIB *gpib )
{
	const char fname[] = "mxi_ni488_device_clear()";

#if 0
	MX_NI488 *ni488;
	int gpib_status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	gpib_status = ibclr( ni488->device_descriptor[address] );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Cannot clear GPIB interface '%s'.  GPIB error = '%s'",
			ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
#else
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"Not yet implemented for NI 488 interfaces." );
#endif
}

MX_EXPORT mx_status_type
mxi_ni488_selective_device_clear( MX_GPIB *gpib, int address )
{
	const char fname[] = "mxi_ni488_selective_device_clear()";

	MX_NI488 *ni488;
	int gpib_status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	gpib_status = ibclr( ni488->device_descriptor[address] );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Cannot clear GPIB address %d on interface '%s'.  GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_local_lockout( MX_GPIB *gpib )
{
	const char fname[] = "mxi_ni488_local_lockout()";

	return mx_error( MXE_UNSUPPORTED, fname,
"The National Instruments GPIB driver does not support 'local lockout'." );
}

MX_EXPORT mx_status_type
mxi_ni488_remote_enable( MX_GPIB *gpib, int address )
{
	const char fname[] = "mxi_ni488_remote_enable()";

	MX_NI488 *ni488;
	int gpib_status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	gpib_status = ibsre( ni488->device_descriptor[address], 1 );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Cannot perform remote enable for GPIB address %d on interface '%s'.  "
	"GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_go_to_local( MX_GPIB *gpib, int address )
{
	const char fname[] = "mxi_ni488_go_to_local()";

#ifdef OS_LINUX
	return mx_error( MXE_UNSUPPORTED, fname,
"The 'go to local' function is not supported by the Linux GPIB driver." );

#else
	MX_NI488 *ni488;
	int gpib_status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	gpib_status = ibloc( ni488->device_descriptor[address] );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Cannot go to local for GPIB address %d on interface '%s'.  GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
#endif
}

MX_EXPORT mx_status_type
mxi_ni488_trigger( MX_GPIB *gpib, int address )
{
	const char fname[] = "mxi_ni488_trigger()";

	MX_NI488 *ni488;
	int gpib_status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	gpib_status = ibtrg( ni488->device_descriptor[address] );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Cannot trigger GPIB address %d on interface '%s'.  GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_wait_for_service_request( MX_GPIB *gpib, double timeout )
{
	const char fname[] = "mxi_ni488_wait_for_service_request()";

	MX_NI488 *ni488;
	int gpib_status, board_address;

	MX_DEBUG( 2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	MX_DEBUG(-2,("%s is not yet implemented correctly.", fname));

	board_address = 0;

	gpib_status = ibwait( ni488->device_descriptor[board_address], 0 );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Wait for service request for GPIB interface '%s' failed.  GPIB error = '%s'",
			ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_serial_poll( MX_GPIB *gpib, int address,
				unsigned char *serial_poll_byte)
{
	const char fname[] = "mxi_ni488_serial_poll()";

	MX_NI488 *ni488;
	int gpib_status;

	MX_DEBUG(-2, ("%s invoked.", fname));

	ni488 = ( MX_NI488 * ) (gpib->record->record_type_struct);

	if ( ni488 == (MX_NI488 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_NI488 pointer for record '%s' is NULL.",
			gpib->record->name );
	}

	MX_DEBUG(-2,
	("%s: must check for correct handling of serial poll byte.",fname ));

	gpib_status = ibrsp( ni488->device_descriptor[address],
					serial_poll_byte );

	if ( gpib_status & ERR ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Serial poll of GPIB address %d on interface '%s' failed.  GPIB error = '%s'",
			address, ni488->interface_name,
			mxi_ni488_gpib_error_text() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_ni488_serial_poll_disable( MX_GPIB *gpib )
{
	/* Apparently, this function is not needed for this interface. */

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_NI488 */
