/*
 * Name:    i_pmac.c
 *
 * Purpose: MX driver for Delta Tau PMAC controllers.  Note that this is
 *          the driver for talking to the PMAC itself.  PMAC-controlled
 *          motors are interfaced via the driver in libmx/d_pmac.c.
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"

#if HAVE_EPICS
#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"
#endif /* HAVE_EPICS */

#include "i_pmac.h"

MX_RECORD_FUNCTION_LIST mxi_pmac_record_function_list = {
	mxi_pmac_initialize_type,
	mxi_pmac_create_record_structures,
	mxi_pmac_finish_record_initialization,
	mxi_pmac_delete_record,
	NULL,
	mxi_pmac_read_parms_from_hardware,
	mxi_pmac_write_parms_to_hardware,
	mxi_pmac_open,
	mxi_pmac_close
};

MX_GENERIC_FUNCTION_LIST mxi_pmac_generic_function_list = {
	mxi_pmac_getchar,
	mxi_pmac_putchar,
	mxi_pmac_read,
	mxi_pmac_write,
	mxi_pmac_input_is_available,
	mxi_pmac_discard_unread_input,
	mxi_pmac_discard_unwritten_output
};

MX_RECORD_FIELD_DEFAULTS mxi_pmac_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_PMAC_STANDARD_FIELDS
};

long mxi_pmac_num_record_fields
		= sizeof( mxi_pmac_record_field_defaults )
			/ sizeof( mxi_pmac_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_pmac_rfield_def_ptr
			= &mxi_pmac_record_field_defaults[0];

#define MXI_PMAC_DEBUG		FALSE

/*==========================*/

MX_EXPORT mx_status_type
mxi_pmac_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_pmac_create_record_structures()";

	MX_GENERIC *generic;
	MX_PMAC *pmac;

	/* Allocate memory for the necessary structures. */

	generic = (MX_GENERIC *) malloc( sizeof(MX_GENERIC) );

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_GENERIC structure." );
	}

	pmac = (MX_PMAC *) malloc( sizeof(MX_PMAC) );

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_PMAC structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = generic;
	record->record_type_struct = pmac;
	record->class_specific_function_list
				= &mxi_pmac_generic_function_list;

	generic->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_pmac_finish_record_initialization()";

	MX_PMAC *pmac;
	char *subtype_name;
	int i, length;

	pmac = (MX_PMAC *) record->record_type_struct;

	/* Identify the subtype of the record. */

	subtype_name = pmac->subtype_name;

	length = strlen( subtype_name );

	for ( i = 0; i < length; i++ ) {
		if ( isupper( subtype_name[i] ) ) {
			subtype_name[i] = tolower( subtype_name[i] );
		}
	}

	if ( strcmp( subtype_name, "rs232" ) == 0 ) {
		pmac->interface_subtype = MX_PMAC_RS232_SUBTYPE;

	} else if ( strcmp( subtype_name, "epics_ect" ) == 0 ) {
		pmac->interface_subtype = MX_PMAC_EPICS_ECT_SUBTYPE;

	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The PMAC interface subtype '%s' is not a valid subtype.",
			subtype_name );
	}

	/* Perform subtype specific processing. */

	switch( pmac->interface_subtype ) {
	case MX_PMAC_RS232_SUBTYPE:
		pmac->interface_record = mx_get_record( record->list_head,
						pmac->interface_args );

		if ( pmac->interface_record == NULL ) {
			return mx_error( MXE_UNDEFINED_RECORD, fname,
			"The RS-232 interface record '%s' does not exist.",
				pmac->interface_args );
		}
		if ( pmac->interface_record->mx_superclass != MXR_INTERFACE ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
				"'%s' is not an interface record.",
				pmac->interface_record->name );
		}
		if ( pmac->interface_record->mx_class != MXI_RS232 ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
				"'%s' is not an RS-232 interface.",
				pmac->interface_record->name );
		}
		break;

	case MX_PMAC_EPICS_ECT_SUBTYPE:
		pmac->interface_record = NULL;
		break;
	}

	/* Mark the PMAC device as being closed. */

	record->flags |= ( ~MXF_REC_OPEN );

	/* Null out the list of attached motors. */

	for ( i = 0; i < MX_NUM_PMAC_MOTORS; i++ ) {
		pmac->motor[i] = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_open( MX_RECORD *record )
{
	const char fname[] = "mxi_pmac_open()";

	MX_PMAC *pmac;
	MX_GENERIC *generic;
	char response[80];
	mx_status_type status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	pmac = (MX_PMAC *) (record->record_type_struct);

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for record '%s' is NULL.", record->name);
	}

	generic = (MX_GENERIC *) pmac->record->record_class_struct;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_GENERIC pointer for record '%s' is NULL.", record->name);
	}

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

#if HAVE_EPICS
	if ( pmac->interface_subtype == MX_PMAC_EPICS_ECT_SUBTYPE ) {
		ezcaAutoErrorMessageOff();
	}
#endif /* HAVE_EPICS */

	/* I haven't yet worked out a foolproof scheme for determining
	 * the value of I3, so I assume the EPICS compatible value of 2.
	 */

	pmac->i3_variable = 2;

	/* Verify that the PMAC controller is active by asking it for
	 * its card type.
	 */

	status = mxi_pmac_putline( pmac, "TYPE", FALSE );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Read the controller type string. */

	status = mxi_pmac_getline(pmac, response, sizeof response, FALSE);

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( strncmp( response, "PMAC", 4 ) != 0 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"The device attached to this interface is not a PMAC controller." );
	}

	/* Now ask for PROM firmware version. */

	status = mxi_pmac_putline( pmac, "VERSION", FALSE );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_pmac_getline( pmac, response, sizeof response, FALSE);

	if ( status.code != MXE_SUCCESS )
		return status;

	sscanf( response, "%d.%d%c",
		&(pmac->major_version),
		&(pmac->minor_version),
		&(pmac->subminor_version));

	status = mxi_pmac_putline( pmac, "I6", FALSE );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_pmac_getline( pmac, response, sizeof response, FALSE);

	if ( status.code != MXE_SUCCESS )
		return status;

	sscanf( response, "%d", &(pmac->i6_variable) );

	MX_DEBUG( 2,
		("%s: PMAC version: major = %d, minor = %d, subminor = %c",
		fname, pmac->major_version, pmac->minor_version,
		pmac->subminor_version));
	MX_DEBUG( 2,("%s: I3 = %d, I6 = %d", fname,
		pmac->i3_variable, pmac->i6_variable));

#if 0
	/* The following is for debugging purposes only. */

	status = mxi_pmac_putline( pmac, "!!!", TRUE );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_pmac_getline( pmac, response, sizeof response, TRUE);

	if ( status.code != MXE_SUCCESS )
		return status;
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_close( MX_RECORD *record )
{
	/* Nothing to do here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_getchar( MX_GENERIC *generic, char *c, int flags )
{
	const char fname[] = "mxi_pmac_getchar()";

	MX_PMAC *pmac;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->interface_subtype ) {
	case MX_PMAC_RS232_SUBTYPE:
		status = mx_rs232_getchar( pmac->interface_record, c, flags );
		break;

	case MX_PMAC_EPICS_ECT_SUBTYPE:
		*c = '\0';

		return mx_error( MXE_UNSUPPORTED, fname,
	"Single character PMAC I/O is not supported for EPICS interfaces." );
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface subtype = %d.",
			pmac->interface_subtype );
		break;
	}
	return status;
}

MX_EXPORT mx_status_type
mxi_pmac_putchar( MX_GENERIC *generic, char c, int flags )
{
	const char fname[] = "mxi_pmac_putchar()";

	MX_PMAC *pmac;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->interface_subtype ) {
	case MX_PMAC_RS232_SUBTYPE:
		status = mx_rs232_putchar( pmac->interface_record, c, flags );
		break;

	case MX_PMAC_EPICS_ECT_SUBTYPE:
		return mx_error( MXE_UNSUPPORTED, fname,
	"Single character PMAC I/O is not supported for EPICS interfaces." );
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface subtype = %d.",
			pmac->interface_subtype );
		break;
	}
	return status;
}

MX_EXPORT mx_status_type
mxi_pmac_read( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_pmac_read()";

	MX_PMAC *pmac;
	MX_RECORD *rs232_record;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->interface_subtype ) {
	case MX_PMAC_RS232_SUBTYPE:
		rs232_record = pmac->interface_record;

		if ( rs232_record == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_RECORD pointer for MX_PMAC structure is NULL." );
		}
		status = mx_rs232_read(rs232_record, buffer, count, NULL, 0);
		break;

	case MX_PMAC_EPICS_ECT_SUBTYPE:
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"The PMAC EPICS driver isn't finished yet.  Sorry." );
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface subtype = %d.",
			pmac->interface_subtype );
		break;
	}
	return status;
}

MX_EXPORT mx_status_type
mxi_pmac_write( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_pmac_write()";

	MX_PMAC *pmac;
	MX_RECORD *rs232_record;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->interface_subtype ) {
	case MX_PMAC_RS232_SUBTYPE:
		rs232_record = pmac->interface_record;

		if ( rs232_record == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_RECORD pointer for MX_PMAC structure is NULL." );
		}
		status = mx_rs232_write(rs232_record, buffer, count, NULL, 0);
		break;

	case MX_PMAC_EPICS_ECT_SUBTYPE:
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"The PMAC EPICS driver isn't finished yet.  Sorry." );
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface subtype = %d.",
			pmac->interface_subtype );
		break;
	}
	return status;
}

MX_EXPORT mx_status_type
mxi_pmac_input_is_available( MX_GENERIC *generic, int *input_available )
{
	const char fname[] = "mxi_pmac_input_is_available()";

	MX_PMAC *pmac;
	MX_RECORD *rs232_record;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->interface_subtype ) {
	case MX_PMAC_RS232_SUBTYPE:
		rs232_record = pmac->interface_record;

		if ( rs232_record == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_RECORD pointer for MX_PMAC structure is NULL." );
		}
		status = mx_rs232_input_is_available(rs232_record,
							input_available);
		break;

	case MX_PMAC_EPICS_ECT_SUBTYPE:
		/* WML: I haven't decided how to implement this yet. */

		*input_available = TRUE;
		status = MX_SUCCESSFUL_RESULT;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface subtype = %d.",
			pmac->interface_subtype );
		break;
	}

	return status;
}

MX_EXPORT mx_status_type
mxi_pmac_discard_unread_input( MX_GENERIC *generic, int debug_flag )
{
	const char fname[] = "mxi_pmac_discard_unread_input()";

	MX_PMAC *pmac;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->interface_subtype ) {
	case MX_PMAC_RS232_SUBTYPE:
		status = mx_rs232_discard_unread_input( pmac->interface_record,
								debug_flag );
		break;

	case MX_PMAC_EPICS_ECT_SUBTYPE:
		/* WML: I haven't decided how to implement this yet. */

		status = MX_SUCCESSFUL_RESULT;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface subtype = %d.",
			pmac->interface_subtype );
		break;
	}
	return status;
}

MX_EXPORT mx_status_type
mxi_pmac_discard_unwritten_output( MX_GENERIC *generic, int debug_flag )
{
	const char fname[] = "mxi_pmac_discard_unwritten_output()";

	MX_PMAC *pmac;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->interface_subtype ) {
	case MX_PMAC_RS232_SUBTYPE:
		status = mx_rs232_discard_unwritten_output(
				pmac->interface_record, debug_flag );
		break;

	case MX_PMAC_EPICS_ECT_SUBTYPE:
		/* As far as I know, this is not possible for EPICS,
		 * so we just say that it succeeded.
		 */

		status = MX_SUCCESSFUL_RESULT;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface subtype = %d.",
			pmac->interface_subtype );
		break;
	}
	return status;
}

/* === Functions specific to this driver. === */

MX_EXPORT mx_status_type
mxi_pmac_command( MX_PMAC *pmac, char *command,
		char *response, int response_buffer_length,
		char *error_message, int error_message_buffer_length,
		int debug_flag )
{
	const char fname[] = "mxi_pmac_command()";

	MX_GENERIC *generic;
	unsigned long sleep_ms;
	int i, max_attempts;
	char alt_error_message_buffer[80];
	char *local_error_buffer;
	int local_error_buffer_length;
	mx_status_type status;

	if ( pmac == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_PMAC pointer passed was NULL." );
	}

	if ( command == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"'command' buffer pointer passed was NULL.  No command sent.");
	}

	if ( error_message == NULL ) {
		local_error_buffer = &alt_error_message_buffer[0];
		local_error_buffer_length = sizeof(alt_error_message_buffer)-1;
	} else {
		local_error_buffer = error_message;
		local_error_buffer_length = error_message_buffer_length;
	}

	if ( local_error_buffer_length < 4 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Error message buffer length of %d is too short.  Must be at least 4.",
			local_error_buffer_length );
	}

	generic = (MX_GENERIC *) pmac->record->record_class_struct;

	/* Send the command string. */

	status = mxi_pmac_putline( pmac, command, debug_flag );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the response, if one is expected. */

	i = 0;

	max_attempts = 10;
	sleep_ms = 1000;

	if ( response != NULL ) {
		for ( i=0; i < max_attempts; i++ ) {
			if ( i > 0 ) {
				MX_DEBUG(-2,
			("%s: No response yet to '%s' command.  Retrying...",
					pmac->record->name, command ));
			}

			status = mxi_pmac_getline( pmac, response,
					response_buffer_length, debug_flag );

			if ( status.code == MXE_SUCCESS ) {
				break;		/* Exit the for() loop. */

			} else if ( status.code != MXE_NOT_READY ) {
				MX_DEBUG(-2,
				("*** Exiting with status = %ld",status.code));
				return status;
			}
			mx_msleep(sleep_ms);
		}

		if ( i >= max_attempts ) {
			status = mxi_pmac_discard_unread_input(
					generic, debug_flag );

			if ( status.code != MXE_SUCCESS ) {
				mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Failed at attempt to discard unread characters in buffer for record '%s'",
					pmac->record->name );
			}

			mx_error( MXE_INTERFACE_IO_ERROR, fname,
				"No response seen to '%s' command", command );
		}
	}
	return status;
}

MX_EXPORT mx_status_type
mxi_pmac_getline( MX_PMAC *pmac,
		char *buffer, long buffer_length, int debug_flag )
{
	const char fname[] = "mxi_pmac_getline()";

	MX_GENERIC *generic;
	char c;
	int i;
	mx_status_type status;

#if HAVE_EPICS
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
#endif /* HAVE_EPICS */

	if ( debug_flag ) {
		MX_DEBUG(-2, ("mxi_pmac_getline() invoked."));
	}

	generic = (MX_GENERIC *) pmac->record->record_class_struct;

	switch( pmac->interface_subtype ) {

		/*========= RS-232 =========*/

	case MX_PMAC_RS232_SUBTYPE:
		/* Depending on the setting of the PMAC I3 variable, there
		 * may be a LF character before the main body of the response.
		 */
		if ( pmac->i3_variable == 1 || pmac->i3_variable == 3 ) {
			status = mx_generic_getchar( generic,
						&c, MXF_GENERIC_WAIT );

			if ( status.code != MXE_SUCCESS ) {
				if ( debug_flag ) {
					MX_DEBUG(-2,
	("%s: did not see any response characters from the PMAC.", fname));
				}
				return status;
			}
			if ( c != MX_LF ) {
				if ( debug_flag ) {
					MX_DEBUG(-2,
	("%s: The first character was not an LF.  Instead saw 0%o", fname, c));
				}
			}
		}
		for ( i = 0; i < buffer_length; i++ ) {
			status = mx_generic_getchar( generic,
						&c, MXF_GENERIC_WAIT );

			if ( status.code != MXE_SUCCESS ) {
				/* Make the buffer contents a valid C string
				 * before returning, so that we can at least
				 * see what appeared before the error.
				 */

				buffer[i] = '\0';

				if ( debug_flag ) {
					MX_DEBUG(-2,
				("mxi_pmac_getline() failed.\nbuffer = '%s'",
						buffer));

					MX_DEBUG(-2,
				("Failed with status = %ld, c = 0x%x '%c'",
					status.code, c, c));
				}
				return status;
			}

			if ( debug_flag ) {
				MX_DEBUG(-2,
			("mxi_pmac_getline: received c = 0x%x '%c'", c, c));
			}

			/* A CR character terminates the response. */

			if ( c == MX_CR ) {
				buffer[i] = '\0';
				break;		/* Exit the for() loop. */
			} else {
				buffer[i] = c;
			}
		}
		if ( i >= buffer_length ) {
			buffer[buffer_length - 1] = '\0';

			status = mx_error( MXE_INTERFACE_IO_ERROR, fname,
				"Warning: Input buffer overrun." );

			/* Exit the switch(pmac->interface_subtype) statement*/
			break;
		}

		/* Depending on the setting of the PMAC I3 variable, there
		 * may be an acknowledgement character after the CR.
		 */

		if ( pmac->i3_variable != 0 ) {

			/* There is an LF or ACK character to throw away. */

			status = mx_generic_getchar( generic,
						&c, MXF_GENERIC_WAIT );

			if ( status.code != MXE_SUCCESS ) {
				if ( debug_flag ) {
					MX_DEBUG(-2,
	("%s: failed to see any character after a CR character.", fname));
				}
				return status;
			}

			switch( pmac->i3_variable ) {
			case 1:
				if ( c != MX_LF ) {
					if ( debug_flag ) {
						MX_DEBUG(-2,
("%s: The character after a CR was not an LF.  Instead saw 0%o", fname, c));
					}
				}
				break;
			case 2:
			case 3:
				if ( c != MX_ACK ) {
					if ( debug_flag ) {
						MX_DEBUG(-2,
("%s: The character after a CR was not an ACK.  Instead saw 0%o", fname, c));
					}
				}
				break;
			}

			if ( debug_flag ) {
				MX_DEBUG(-2,
			("mxi_pmac_getline: received c = 0x%x '%c'", c, c));
			}
		}
		break;    /* End of case MX_PMAC_RS232_SUBTYPE */

		/*========= EPICS =========*/

	case MX_PMAC_EPICS_ECT_SUBTYPE:
#if HAVE_EPICS
		sprintf( pvname, "%sStrRsp", pmac->interface_args );

		ezca_status = ezcaGet( pvname, ezcaString, 1, buffer );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error receiving string from EPICS variable '%s'.  EZCA status = %d (%s)",
			pvname, ezca_status, ezca_error_message );

			return mx_error( MXE_DEVICE_IO_ERROR, fname,
				mx_error_message );
		}
#endif /* HAVE_EPICS */
		break;

		/*========= default =========*/

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface subtype = %d.",
			pmac->interface_subtype );
		break;
	}

#if MXI_PMAC_DEBUG
	if ( 1 ) {
#else
	if ( debug_flag ) {
#endif
		MX_DEBUG(-2, ("mxi_pmac_getline: buffer = '%s'", buffer) );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_putline( MX_PMAC *pmac, char *buffer, int debug_flag )
{
	const char fname[] = "mxi_pmac_putline()";

	MX_GENERIC *generic;
	char *ptr;
	char c;
	mx_status_type status;

#if HAVE_EPICS
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
#endif /* HAVE_EPICS */

#if MXI_PMAC_DEBUG
	if ( 1 ) {
#else
	if ( debug_flag ) {
#endif
		MX_DEBUG(-2, ("mxi_pmac_putline: buffer = '%s'",buffer));
	}

	generic = (MX_GENERIC *) pmac->record->record_class_struct;

	switch( pmac->interface_subtype ) {
	case MX_PMAC_RS232_SUBTYPE:
		/* First send a ctrl-Z to make the serial port the
		 * active interface.
		 */

		c = MX_CTRL_Z;

		status = mx_generic_putchar( generic, c, MXF_GENERIC_WAIT );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( debug_flag ) {
			MX_DEBUG(-2,("mxi_pmac_putline: sent 0x%x '%c'",c,c));
		}

		/* Now send the main string. */

		ptr = &buffer[0];

		while ( *ptr != '\0' ) {
			c = *ptr;

			status = mx_generic_putchar( generic,
					c, MXF_GENERIC_WAIT );

			if ( status.code != MXE_SUCCESS )
				return status;

			if ( debug_flag ) {
				MX_DEBUG(-2,
				("mxi_pmac_putline: sent 0x%x '%c'", c, c));
			}
			ptr++;
		}
		/* Send a CR after the end of the command. */

		c = MX_CR;

		status = mx_generic_putchar( generic, c, MXF_GENERIC_WAIT );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( debug_flag ) {
			MX_DEBUG(-2,(
			"mxi_pmac_putline(eos): sent 0x%x '%c'", c, c));
		}

		/* Check for an acknowledgement character from the PMAC. */

		status = mx_generic_getchar( generic, &c, MXF_GENERIC_WAIT );

		if ( status.code != MXE_SUCCESS ) {
			/* For I3 = 0, not returning a character is the
			 * normal response.
			 */
			if ( pmac->i3_variable == 0 ) {
				/* Exit the switch( pmac->interface_subtype )
				 * statement.
				 */
				break;
			}
			if ( debug_flag ) {
				MX_DEBUG(-2,
	("%s: failed to see any character after a CR character.", fname));
			}
			return status;
		}

		if ( c == MX_BELL ) {
			MX_DEBUG(-2,
("%s: BELL character seen.  This means the PMAC controller reported an error.",
				fname));
		}

		switch( pmac->i3_variable ) {
		case 1:
			if ( c != MX_LF ) {
				if ( debug_flag ) {
					MX_DEBUG(-2,
("%s: The character after a CR was not an LF.  Instead saw 0%o", fname, c));
				}
			}
			break;
		case 2:
		case 3:
			if ( c != MX_ACK ) {
				if ( debug_flag ) {
					MX_DEBUG(-2,
("%s: The character after a CR was not an ACK.  Instead saw 0%o", fname, c));
				}
			}
			break;
		}

		if ( debug_flag ) {
			MX_DEBUG(-2,
			("mxi_pmac_putline: received c = 0x%x '%c'", c, c));
		}
		break;

	case MX_PMAC_EPICS_ECT_SUBTYPE:
#if HAVE_EPICS
		sprintf( pvname, "%sStrCmd", pmac->interface_args );

		ezca_status = ezcaPut( pvname, ezcaString, 1, buffer );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error sending string '%s' to EPICS variable '%s'.  EZCA status = %d (%s)",
			buffer, pvname, ezca_status, ezca_error_message );

			return mx_error( MXE_DEVICE_IO_ERROR, fname,
				mx_error_message );
		}
#endif /* HAVE_EPICS */
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface subtype = %d.",
			pmac->interface_subtype );
		break;
	}
	return MX_SUCCESSFUL_RESULT;
}

