/*
 * Name:    mx_encod.c
 *
 * Purpose: MX function library of generic encoder operations.
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_encod.h"

MX_EXPORT mx_status_type
mx_encoder_get_pointers( MX_RECORD *encoder_record,
			MX_ENCODER **encoder,
			MX_ENCODER_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_encoder_get_pointers()";

	if ( encoder_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The encoder_record pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( encoder_record->mx_class != MXC_ENCODER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The record '%s' passed by '%s' is not a ENCODER interface.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			encoder_record->name, calling_fname,
			encoder_record->mx_superclass,
			encoder_record->mx_class,
			encoder_record->mx_type );
	}

	if ( encoder != (MX_ENCODER **) NULL ) {
		*encoder = (MX_ENCODER *) (encoder_record->record_class_struct);

		if ( *encoder == (MX_ENCODER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_ENCODER pointer for record '%s' passed by '%s' is NULL.",
				encoder_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_ENCODER_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_ENCODER_FUNCTION_LIST *)
				(encoder_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_ENCODER_FUNCTION_LIST *) NULL )
		{
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_ENCODER_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				encoder_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_encoder_get_limit_status( MX_RECORD *encoder_record,
		int *underflow_set, int *overflow_set )
{
	const char fname[] = "mx_encoder_get_limit_status()";

	MX_ENCODER *encoder;
	MX_ENCODER_FUNCTION_LIST *function_list;
	mx_status_type ( *get_limit_status_fn ) ( MX_ENCODER *, int *, int * );
	mx_status_type status;

	status = mx_encoder_get_pointers( encoder_record, &encoder,
					&function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_limit_status_fn = function_list->get_limit_status;

	if ( get_limit_status_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"overflow_set function ptr for MX_ENCODER ptr 0x%p is NULL.",
			encoder);
	}

	/* All is well, so invoke the function. */

	status = (*get_limit_status_fn)(encoder, underflow_set, overflow_set);

	return status;
}

MX_EXPORT mx_status_type
mx_encoder_reset_limit_status( MX_RECORD *encoder_record )
{
	const char fname[] = "mx_encoder_reset_limit_status()";

	MX_ENCODER *encoder;
	MX_ENCODER_FUNCTION_LIST *function_list;
	mx_status_type ( *reset_limit_status_fn ) ( MX_ENCODER * );
	mx_status_type status;

	status = mx_encoder_get_pointers( encoder_record, &encoder,
					&function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	reset_limit_status_fn = function_list->reset_limit_status;

	if ( reset_limit_status_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"reset_limit_status function ptr for MX_ENCODER ptr 0x%p is NULL.",
			encoder);
	}

	/* All is well, so invoke the function. */

	status = (*reset_limit_status_fn)( encoder );

	return status;
}

MX_EXPORT mx_status_type
mx_encoder_read( MX_RECORD *encoder_record, long *value )
{
	const char fname[] = "mx_encoder_read()";

	MX_ENCODER *encoder;
	MX_ENCODER_FUNCTION_LIST *function_list;
	mx_status_type ( *read_fn ) ( MX_ENCODER *, long * );
	mx_status_type status;

	status = mx_encoder_get_pointers( encoder_record, &encoder,
					&function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	read_fn = function_list->read;

	if ( read_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for MX_ENCODER ptr 0x%p is NULL.",
			encoder);
	}

	/* All is well, so invoke the function. */

	status = (*read_fn)( encoder, value );

	return status;
}

MX_EXPORT mx_status_type
mx_encoder_write( MX_RECORD *encoder_record, long value )
{
	const char fname[] = "mx_encoder_write()";

	MX_ENCODER *encoder;
	MX_ENCODER_FUNCTION_LIST *function_list;
	mx_status_type ( *write_fn ) ( MX_ENCODER *, long );
	mx_status_type status;

	status = mx_encoder_get_pointers( encoder_record, &encoder,
					&function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( encoder->encoder_flags & MXF_ENC_ABSOLUTE_ENCODER ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Writing to an absolute encoder is not allowed.");
	}

	write_fn = function_list->write;

	if ( write_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for MX_ENCODER ptr 0x%p is NULL.",
			encoder);
	}

	/* All is well, so invoke the function. */

	status = (*write_fn)( encoder, value );

	return status;
}

