/*
 * Name:    mx_var.h
 *
 * Purpose: Header file for generic variable superclass support.
 *
 * Warning: The prototypes for the 1-D variable functions are scheduled to
 *          be changed sometime during 2000.  The changes will primarily
 *          be to replace arguments that pass string names and record list
 *          pointers with record pointers.  For example, mx_get_long_variable()
 *          will probably be changed to something like:
 *
 *          MX_API mx_status_type mx_get_long_variable(
 *                                      MX_RECORD *variable_record,
 *                                      long *long_value );
 *
 *          The current version will probably be renamed to something like
 *          mx_get_long_variable_by_name().
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */


#ifndef _MX_VAR_H_
#define _MX_VAR_H_

typedef struct {
	MX_RECORD *record; /* Pointer to the MX_RECORD structure that points
			    * to this MX_SCAN structure.
			    */
	long num_dimensions;
	long *dimension;

	void *pointer_to_value;
} MX_VARIABLE;

typedef struct {
	mx_status_type ( *send_value )( MX_VARIABLE *variable );
	mx_status_type ( *receive_value )( MX_VARIABLE *variable );
} MX_VARIABLE_FUNCTION_LIST;

MX_API_PRIVATE mx_status_type mx_variable_initialize_type( long record_type );

MX_API_PRIVATE mx_status_type
		mx_variable_fixup_varargs_record_field_defaults(
		MX_RECORD_FIELD_DEFAULTS *record_field_defaults_array,
		long num_record_fields );

MX_API mx_status_type mx_send_variable( MX_RECORD *record );
MX_API mx_status_type mx_receive_variable( MX_RECORD *record );

/* The following functions can be regarded to some extent as stopgaps
 * until the client/server functions can be written.
 */

MX_API mx_status_type mx_get_variable_parameters( MX_RECORD *record,
	long *num_dimensions, long **dimension_array, long *field_type );

MX_API mx_status_type mx_get_variable_pointer( MX_RECORD *record,
						void **pointer_to_value );

/* Note that the following functions only work for 1-dimensional arrays. */

MX_API mx_status_type mx_get_variable_with_type( MX_RECORD *record_list,
			char *record_name, long field_type,
			long *num_elements, void **pointer_to_value );

MX_API mx_status_type mx_get_variable_from_record( MX_RECORD *record,
				long field_type, long *num_elements,
				void **pointer_to_value );

MX_API mx_status_type mx_get_long_variable( MX_RECORD *record_list,
		char *record_name, long *long_value );

MX_API mx_status_type mx_get_unsigned_long_variable( MX_RECORD *record_list,
		char *record_name, unsigned long *unsigned_long_value );

MX_API mx_status_type mx_get_double_variable( MX_RECORD *record_list,
		char *record_name, double *double_value );

MX_API mx_status_type mx_get_string_variable( MX_RECORD *record_list,
	char *record_name, char *string_value, size_t max_string_length );

/* The following definitions of MXA_XXX_SIZEOF assume that
 * MXU_FIELD_MAX_DIMENSIONS is equal to 8.
 */

#define MXA_STRING_SIZEOF \
	{ sizeof(char), sizeof(char *), \
	sizeof(char **), sizeof(char ***), \
	sizeof(char ****), sizeof(char *****), \
	sizeof(char ******), sizeof(char *******) }

#define MXA_CHAR_SIZEOF     MXA_STRING_SIZEOF

#define MXA_UCHAR_SIZEOF \
	{ sizeof(unsigned char), sizeof(unsigned char *), \
	sizeof(unsigned char **), sizeof(unsigned char ***), \
	sizeof(unsigned char ****), sizeof(unsigned char *****), \
	sizeof(unsigned char ******), sizeof(unsigned char *******) }

#define MXA_SHORT_SIZEOF \
	{ sizeof(short), sizeof(short *), \
	sizeof(short **), sizeof(short ***), \
	sizeof(short ****), sizeof(short *****), \
	sizeof(short ******), sizeof(short *******) }

#define MXA_USHORT_SIZEOF \
	{ sizeof(unsigned short), sizeof(unsigned short *), \
	sizeof(unsigned short **), sizeof(unsigned short ***), \
	sizeof(unsigned short ****), sizeof(unsigned short *****), \
	sizeof(unsigned short ******), sizeof(unsigned short *******) }

#define MXA_INT_SIZEOF \
	{ sizeof(int), sizeof(int *), \
	sizeof(int **), sizeof(int ***), \
	sizeof(int ****), sizeof(int *****), \
	sizeof(int ******), sizeof(int *******) }

#define MXA_UINT_SIZEOF \
	{ sizeof(unsigned int), sizeof(unsigned int *), \
	sizeof(unsigned int **), sizeof(unsigned int ***), \
	sizeof(unsigned int ****), sizeof(unsigned int *****), \
	sizeof(unsigned int ******), sizeof(unsigned int *******) }

#define MXA_LONG_SIZEOF \
	{ sizeof(long), sizeof(long *), \
	sizeof(long **), sizeof(long ***), \
	sizeof(long ****), sizeof(long *****), \
	sizeof(long ******), sizeof(long *******) }

#define MXA_ULONG_SIZEOF \
	{ sizeof(unsigned long), sizeof(unsigned long *), \
	sizeof(unsigned long **), sizeof(unsigned long ***), \
	sizeof(unsigned long ****), sizeof(unsigned long *****), \
	sizeof(unsigned long ******), sizeof(unsigned long *******) }

#define MXA_FLOAT_SIZEOF \
	{ sizeof(float), sizeof(float *), \
	sizeof(float **), sizeof(float ***), \
	sizeof(float ****), sizeof(float *****), \
	sizeof(float ******), sizeof(float *******) }

#define MXA_DOUBLE_SIZEOF \
	{ sizeof(double), sizeof(double *), \
	sizeof(double **), sizeof(double ***), \
	sizeof(double ****), sizeof(double *****), \
	sizeof(double ******), sizeof(double *******) }

#define MXA_RECORD_SIZEOF \
	{ sizeof(MX_RECORD *), sizeof(MX_RECORD **), \
	sizeof(MX_RECORD ***), sizeof(MX_RECORD ****), \
	sizeof(MX_RECORD *****), sizeof(MX_RECORD ******), \
	sizeof(MX_RECORD *******), sizeof(MX_RECORD ********) }

/* =========================================== */

#define MX_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "num_dimensions", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, num_dimensions), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
\
  {-1, -1, "dimension", MXFT_LONG, NULL, 1, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, dimension), \
	{sizeof(long)}, NULL, (MXFF_IN_DESCRIPTION | MXFF_VARARGS) }


/* =========================================== */

#define MX_STRING_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_STRING, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_STRING_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_CHAR_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_CHAR, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_CHAR_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_UCHAR_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_UCHAR, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_UCHAR_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_SHORT_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_SHORT, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_SHORT_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_USHORT_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_USHORT, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_USHORT_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_INT_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_INT, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_INT_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_UINT_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_UINT, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_UINT_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_LONG_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_LONG, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_LONG_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_ULONG_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_ULONG, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_ULONG_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_FLOAT_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_FLOAT, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_FLOAT_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_DOUBLE_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_DOUBLE, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_DOUBLE_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#define MX_RECORD_VARIABLE_STANDARD_FIELDS \
  {-1, -1, "value", MXFT_RECORD, NULL, \
	MXU_VARARGS_LENGTH, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_VARIABLE, pointer_to_value), \
	MXA_RECORD_SIZEOF, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS) }

#endif /* _MX_VAR_H_ */

